// Stand Alone Game Checksum Fixer
// (for The Games Factory, Click 'n' Create and Multimedia Fusion Express)

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#ifdef WIN32
#include <conio.h>
#endif

int main(int argc, char* argv[]);
unsigned char FixCheckSum(char* FileName);
unsigned short int CalcCheckSum(unsigned char Mode, unsigned long int DataLen,
								unsigned char* Data);
unsigned long int SearchForData(unsigned long int DataLen, unsigned char* Data,
								unsigned long int MagicVal);

int main(int argc, char* argv[])
{
	unsigned char RetVal;
	
	printf("TGF/CnC Game Checksum Fixer\n");
	printf("---------------------------\n");
	printf("\n");
	
	if (argc <= 1 || ! stricmp(argv[0x01], "-help"))
	{
		printf("Usage: GameChkS.exe GameName\n");
		printf("GameName  Game Executable (with or without extention)\n");
		return 0;
	}
	
	RetVal = FixCheckSum(argv[0x01]);
	
#ifdef WIN32
	if (argv[0][1] == ':')
	{
		// Executed by Double-Clicking (or Drap and Drop)
		if (_kbhit())
			_getch();
		_getch();
	}
#endif
	
	return RetVal;
}

unsigned char FixCheckSum(char* FileName)
{
	char* EXEName;
	char* GameName;
	FILE* hEFile;
	FILE* hGFile;
	unsigned long int GameDataLen;
	unsigned char* GameData;
	unsigned long int EXEDataLen;
	unsigned char* EXEData;
	unsigned long int CurPos;
	unsigned short int* Int16Ptr;
	unsigned char GameType;
	const char* GamTName;
	char* StrPnt;
	
	// Make EXE Name
	EXEName = (char*)malloc(strlen(FileName) + 0x08);
	strcpy(EXEName, FileName);
	StrPnt = strrchr(EXEName, '.');
	if (StrPnt == NULL || stricmp(StrPnt, ".exe"))
		strcat(EXEName, ".exe");
	
	
	// Load EXE File
	hEFile = fopen(EXEName, "r+b");
	if (hEFile == NULL)
	{
		printf("Error opening %s!", EXEName);
		return 1;
	}
	
	fseek(hEFile, 0x00, SEEK_END);
	EXEDataLen = ftell(hEFile);
	if (EXEDataLen > 0x80000)
		EXEDataLen = 0x80000;	// limit size to 512 KB
	
	EXEData = (unsigned char*)malloc(EXEDataLen);
	if (EXEData == NULL)
	{
		printf("Not enough memory!\n");
		fclose(hEFile);
		return 4;
	}
	fseek(hEFile, 0x00, SEEK_SET);
	fread(EXEData, 0x01, EXEDataLen, hEFile);
	printf("EXE opened.\n");
	// File is closed after writing the checksum
	
	
	// Check for EXE file
	Int16Ptr = (unsigned short int*)&EXEData[0x00];
	if (*Int16Ptr != 0x5A4D)	// 4D5A - MZ
	{
		printf("That's no EXE file. What do you want??\n");
		free(EXEData);
		fclose(hEFile);
		return 2;
	}
	
	// Search for Checksum (also check if the file is valid)
	CurPos = SearchForData(EXEDataLen, EXEData, 0x56458978);
	if (CurPos == 0xFFFFFFFF)
	{
		printf("Can't find Checksum. Please choose a Stand Alone Game!\n");
		free(EXEData);
		fclose(hEFile);
		return 2;
	}
	Int16Ptr = (unsigned short int*)&EXEData[CurPos];
	
	// get extention of Game Data File
	CurPos = SearchForData(EXEDataLen, EXEData, 0x01024544);
	if (CurPos == 0xFFFFFFFF)
	{
		GameType = 0x00;
	}
	else
	{
		// That's as much of EXE-Data as we need
		memcpy(&EXEDataLen, &EXEData[CurPos + 0x00], 0x04);
		
		// search the extention of the data file (needs a little work,
		// because of slight differences between 16-bit and 32-bit versions)
		StrPnt = &EXEData[CurPos + 0x06];
		while(! *StrPnt)
			StrPnt ++;
		StrPnt += 0x44;
		while(! *StrPnt)
			StrPnt ++;
		
		if (! strcmp(StrPnt, ".gam"))
		{
			GameType = 0x01;	// TGF and gam
			GamTName = "The Games Factory";
		}
		else if (! strcmp(StrPnt, ".cca"))
		{
			GameType = 0x02;	// CnC/MMFx and cca
			GamTName = "Click 'n' Create / Multimedia Fusion Express";
		}
		else
			GameType = 0x00;
	}
	if (! GameType)
	{
		printf("Unable to detect Game Type!\n");
		free(EXEData);
		fclose(hEFile);
		return 3;
	}
	printf("Game Type:\n\t%s (%s)\n", GamTName, StrPnt);
	
	
	// Load Game File
	GamTName = StrPnt;
	GameName = (char*)malloc(strlen(EXEName) + 0x01);
	strcpy(GameName, EXEName);
	StrPnt = strrchr(GameName, '.');
	if (StrPnt == NULL)
	{
		printf("Error making Game Filename! (This really shouldn't happen.)\n");
		free(EXEData);
		fclose(hEFile);
		return 0;
	}
	strcpy(StrPnt, GamTName);
	
	hGFile = fopen(GameName, "rb");
	if (hGFile == NULL)
	{
		printf("Error opening %s!", GameName);
		free(EXEData);
		fclose(hEFile);
		return 1;
	}
	
	fseek(hGFile, 0x00, SEEK_END);
	GameDataLen = ftell(hGFile);
	if (GameDataLen > 0x200)
		GameDataLen = 0x200;	// limit size to 512 Bytes (we don't need more)
	
	GameData = (unsigned char*)malloc(GameDataLen);
	if (GameData == NULL)
	{
		printf("Not enough memory!\n");
		free(EXEData);
		fclose(hEFile);
		fclose(hGFile);
		return 4;
	}
	fseek(hGFile, 0x00, SEEK_SET);
	fread(GameData, 0x01, GameDataLen, hGFile);
	printf("%s loaded.\n", GamTName);
	
	fclose(hGFile);
	

	// calculate and write new Game Checksum
	printf("Game Checksum:\n");
	printf("\tOld:\t%04hX\n", *Int16Ptr);
	*Int16Ptr = CalcCheckSum(0x00, GameDataLen, GameData);
	printf("\tNew:\t%04hX\n", *Int16Ptr);
	free(GameData);
	
	// Search EXE Checksum
	CurPos = SearchForData(EXEDataLen, EXEData, 0x58495216);
	printf("EXE Checksum:\n");
	if (CurPos == 0xFFFFFFFF)
	{
		// old TGF/Cnc/MMFx games lack a EXE checksum
		printf("\tmissing\n");
	}
	else
	{
		// calculate and write new EXE Checksum
		Int16Ptr = (unsigned short int*)&EXEData[CurPos];
		printf("\tOld:\t%04hX\n", *Int16Ptr);
		*Int16Ptr = CalcCheckSum(0x01, EXEDataLen, EXEData);
		printf("\tNew:\t%04hX\n", *Int16Ptr);
	}
	

	// Write EXE File
	fseek(hEFile, 0x00, SEEK_SET);
	fwrite(EXEData, 0x01, EXEDataLen, hEFile);
	printf("EXE written.\n");
	
	fclose(hEFile);
	
	return 0;
}

unsigned short int CalcCheckSum(unsigned char Mode, unsigned long int DataLen,
								unsigned char* Data)
{
	// Warning! CalcCheckSum will WRITE to Data to mask out values!
	unsigned long int DataEnd;
	unsigned char* Buffer;
	unsigned short int* Int16Ptr;
	unsigned long int CurPos;
	unsigned short int ChkSum;
	
	switch(Mode)
	{
	case 0x00:	// gam or cca-file
		memcpy(&ChkSum, &Data[0x15C], 0x02);
		if (ChkSum > 0x20)
			ChkSum = 0x20;
		DataEnd = 0x0164 + ChkSum * 0x04;
		
		Buffer = Data + 0x02;
		Int16Ptr = (unsigned short int*)&Buffer[0xF2];
		*Int16Ptr = 0x0000;
		Int16Ptr = (unsigned short int*)&Buffer[0xFE];
		*Int16Ptr &= 0x1FFF;
		
		break;
	case 0x01:	// exe-file
		Buffer = Data;
		CurPos = SearchForData(DataLen, Buffer, 0x58495216);
		if (CurPos == 0xFFFFFFFF)
			return 0x0000;
		Int16Ptr = (unsigned short int*)&Buffer[CurPos];
		*Int16Ptr = 0x0000;
		
		CurPos = SearchForData(DataLen, Data, 0x01024544);
		if (CurPos == 0xFFFFFFFF)
		{
			printf("Error reading EXE size from the game! Trying checksum over complete file!\n");
			DataEnd = DataLen;
		}
		else
		{
			memcpy(&DataEnd, &Buffer[CurPos], 0x04);
		}
		break;
	}
	
	ChkSum = 0x00;
	for (CurPos = 0x00; CurPos < DataEnd; CurPos ++)
	{
		ChkSum += Buffer[CurPos];
		// 16-bit rotl(ChkSum, 1)
		ChkSum = (ChkSum << 1) | ((ChkSum & 0x8000) >> 15);
	}
	
	return ChkSum;
}

unsigned long int SearchForData(unsigned long int DataLen, unsigned char* Data,
								unsigned long int MagicVal)
{
	unsigned long int CurPos;
	unsigned short int* Int16Pnt;
	unsigned long int* Int32Pnt;
	
	// known Data Keywords:
	//	TTM
	//	12
	//	 X 	7777 1652 4958 - EXE Checksum [WORD]
	//	XX 	7777 7889 4556 - Game Checksum [WORD]
	//	XX 	7777 0335 2478 - ???
	//	XX 	7777 8715 3030 - ???
	//	XX 	7777 2468 ACE0 - ???
	//	XXX	7777 4445 0201 - Extention Pos [DWORD] (also EXE Size for Checksum) and
	//		                  number of attached extentions [WORD]
	//	XX 	7777 5994 3972 - ???
	//	  X	7777 4987 4712 - ??? (MMF only)
	//	T - TGF/CnC/MMF Express (1 - 16 bit, 2 - 32 bit)
	//	M - MMF
	Int16Pnt = (unsigned short int*)Data;
	for (CurPos = 0x00; CurPos < DataLen - 0x05; CurPos += 0x02, Int16Pnt ++)
	{
		if (*Int16Pnt == 0x7777)
		{
			Int32Pnt = (unsigned long int*)&Data[CurPos + 0x02];
			if (*Int32Pnt == MagicVal)
				return CurPos + 0x06;
		}
	}
	
	return 0xFFFFFFFF;
}
